//==========================================================================
// Implementacia suboroveho systemu FAT16 s vyuzitim SD/MMC karty. 
// Created By : Nader Chini Chian 
// Email : n...@yahoo.com 
// Upravene a prelozene : Ondrej Karpis
//========================================================================== 

#include "sd_spi.h"	// programy pre komunikaciu s SD kartou cez SPI
#include "retazec.h"	// programy pre pracu s retazcami

#define MasterBootRecordAdd 0 
#define BPB_BytsPerSec 512
//Volume Boot Record location in Master Boot Record
#define VBR_Addr 0x1C6
//define DIR_Attr
#define ATTR_LONG_NAME 0x0F
#define ATTR_READ_ONLY 0x01
#define ATTR_HIDDEN 0x02
#define ATTR_SYSTEM 0x04
#define ATTR_VOLUME_ID 0x08
#define ATTR_DIRECTORY 0x10
#define ATTR_ARCHIVE 0x20
// For working with buffer syncronously
#define FULL 1 
#define EMPTY 0 
#define FILE_NOT_FOUND 255 

//===========================================================================
// Struktura pre pracu s vlastnostami subora 
//===========================================================================
typedef struct 
{ unsigned char DIR_Name[11]; // 11 char file name 
  unsigned char DIR_Attr; //file attributes RSHA, Longname, Drive Label, Directory
  unsigned char DIR_NTRes; //set to zero
  unsigned char DIR_CrtTimeTenth; //creation time part in milliseconds
  short DIR_CrtTime; //creation time
  short DIR_CrtDate; //creation date
  short DIR_LastAccDate; //last access date
  short DIR_FstClusHI; //first cluster high word 
  short DIR_WrtTime; //last write time
  short DIR_WrtDate; //last write date
  short DIR_FstClusLO; //first cluster low word 
  unsigned int DIR_FileSize; 
}DIR_ENTRY; 

short file_time;
short file_date;

//===========================================================================
// Union pre nacitanie poloziek adresara
//===========================================================================
unsigned int BufferStatus1 = EMPTY;
unsigned int BufferStatus2 = EMPTY;
union BUFFERWORK1
{ unsigned char data[512];
  DIR_ENTRY DirEntry[16]; 
} Buffer1; 

//===========================================================================
// Union pre ulozenie jedneho sektoru FAT
//=========================================================================== 
union FATDataStruct
{ unsigned char Buffer[512];
  short Cluster[256];
} FAT; 

//===========================================================================
// Premenne pre pracu s FAT
//===========================================================================
unsigned int LastFATSectorRead = 0; 
unsigned char LastFreeClusSector; 
unsigned char SETCLUS = 0; // for speed up the GetClusEntryVal_N func 

//===========================================================================
unsigned int BootSectorAdd; 
unsigned int FATAdd;
unsigned int RootDirAdd;
unsigned int FirstDataSector; 
unsigned char RootDirSectors; 
unsigned char BPB_SecPerClus;
unsigned int BPB_RsvdSecCnt; 
unsigned char BPB_NumFATs;
unsigned int BPB_RootEntCnt; 
unsigned int BPB_TotSec16;
unsigned int BPB_FATSz16; 
unsigned int BPB_TotSec32; 


//===========================================================================
// Pomocna funkcia na upravu mena subora (z "pokus.txt" urobi "POKUS   TXT")
//===========================================================================
void uprav_meno(unsigned char *s, unsigned char *d)
{ int i, j;
  unsigned char ch;

  // meno subora je maximalne 8 znakov plus bodka plus 3 znaky pripona
  for(i=0,j=0;i<12;i++,j++)
  { ch = s[i];
    if (ch == 0)
      break;
    if (ch == '.')
    { // doplnenie medzerami do 8 znakov
      for(;j<8;j++)
        d[j] = ' ';
      // odstranenie dvojnasobneho zvysenia j  
      j--;  
    }
    else
    { // uprava malych pismen na velke
      if (ch >= 'a' && ch <= 'z')
        d[j] = ch - 32;
       else 
        d[j] = ch;
    }    
  }
  // doplnenie medzier ak treba
  for(;j<12;j++)
   d[j] = ' ';
}

//===========================================================================
// Inicializacia FAT16, treba zavolat na zaciatku prace s FAT
//===========================================================================
void FAT_Init(unsigned char *Buffer)
{
  //Read Master Boot Record
  SD_Read(MasterBootRecordAdd,Buffer); 
//  BootSectorAdd = *( (unsigned int *) &Buffer[VBR_Addr] ); 
  BootSectorAdd = 0;
  //Read Boot Sector
  SD_Read(BootSectorAdd,Buffer); 
  //Parse the boot sector
  BPB_SecPerClus = Buffer[13];
  BPB_RsvdSecCnt = Buffer[14] + 256 * Buffer[15];
  BPB_NumFATs = Buffer[16];
  BPB_RootEntCnt = Buffer[17] + 256 * Buffer[18];
  BPB_TotSec16 = Buffer[19] + 256 * Buffer[20];
  BPB_FATSz16 = Buffer[22] + 256 * Buffer[23];
  BPB_TotSec32 = Buffer[32] + 256 * Buffer[33] + 65536*Buffer[34] + 256* 65536*Buffer[35];
  FATAdd = BootSectorAdd + BPB_RsvdSecCnt;
  RootDirAdd = FATAdd + (BPB_NumFATs * BPB_FATSz16);
  //determine the count of sectors occupied by root directory
  RootDirSectors = ((BPB_RootEntCnt * 32) + (BPB_BytsPerSec - 1)) / BPB_BytsPerSec;
  FirstDataSector = RootDirAdd + RootDirSectors; 
  // for speed up for first time call FindFreeClus
  LastFreeClusSector = FATAdd; 
} 

//===========================================================================
// Podprogramy pre pracu s clustermi
//=========================================================================== 
unsigned int FirstSectorofCluster_N(unsigned int N)
{
  return( ((N - 2) * BPB_SecPerClus) + FirstDataSector );
}

//=========================================================================== 
short GetClusEntryVal_N(unsigned int N)
{ // test if the last sector read of fat table != needed sector so update the previous sector in mmc 
  // card & read new sector from mmc card & save it in memory. 
  if(LastFATSectorRead != FATAdd + ( N >> 8 )) 
  { // if until now we didn't read any sector so we don't need any update 
    // if we use this function continously like in ReadFile so don't need any update except when
    // using SetClusEntryVal_N before it(GetClusEntryVal_N)
    if( LastFATSectorRead != 0 && SETCLUS == 1)
    { 
      // sector fat table needed now change so update the fat table
      SD_Write( LastFATSectorRead , FAT.Buffer); 
      SETCLUS = 0;
    } 
    // find the sector number of fat needed & read it in memory
    SD_Read( FATAdd + ( N >> 8 ) , FAT.Buffer);
    // set the lastSectorRead with the new sector readed
    LastFATSectorRead = FATAdd + ( N >> 8 );
  } 
  // find & return the entry of the cluster number N that in memory
  return(FAT.Cluster[ N % 256 ] );
} 

//===========================================================================
void SetClusEntryVal_N(unsigned int N, unsigned int value)
{ // test if the last sector read of fat table != needed sector so update the previous sector in mmc 
  // card & read new sector from mmc card & save it in memory. 
  if(LastFATSectorRead != FATAdd + ( N >> 8 )) 
  { // if until now we didn't read any sector so we don't need any update
    if( LastFATSectorRead != 0) 
    // sector fat table needed now change so update the fat table 
    SD_Write( LastFATSectorRead , FAT.Buffer); 
    // find the sector number of fat needed & read it in memory 
    SD_Read( FATAdd + ( N >> 8 ) , FAT.Buffer); 
    // set the lastSectorRead with the new sector readed 
    LastFATSectorRead = FATAdd + ( N >> 8 ); 
  } 
  // set the cluster entry value with new value 
  FAT.Cluster[ N % 256 ] = value; 
  SETCLUS = 1;
} 

//===========================================================================
void UpdateFATBuffer(void)
{
  if(LastFATSectorRead >= FATAdd )
    SD_Write( LastFATSectorRead , FAT.Buffer); 
}

//===========================================================================
unsigned int FindFreeCluster(void)
{ unsigned char j , UPDATE = 0 , SEARCH_COMPLETE = 0; 

  // See if the sector readed is equal to LastFreeClusSector if not update the FAT table & read it
  if(LastFreeClusSector != LastFATSectorRead)
  { // if until now we didn't read any sector so we don't need any update 

    if( LastFATSectorRead != 0) 
      // sector fat table needed now change so update the fat table 
      SD_Write( LastFATSectorRead , FAT.Buffer); 
    // find the sector number of fat needed & read it in memory 
    SD_Read( LastFreeClusSector , FAT.Buffer); 
    // set the lastSectorRead with the new sector readed 
    LastFATSectorRead = LastFreeClusSector; 
    UPDATE = 1; 
  } 
  // This loop search total of sectors of the FAT Table until find a free cluster
  while( LastFreeClusSector < (FATAdd + BPB_FATSz16) )
  { // search for free cluster on a sector that readed before
    j = 0; 
    do { if(FAT.Cluster[ j ] == 0x0000) 
         { LastFATSectorRead = LastFreeClusSector; 
           FAT.Cluster[ j ] = 0xffff;
           return( (LastFreeClusSector - FATAdd)*256 + j ); 
         } 
       } 
    while(j++ != 255); 

    // if until now we didn't update fat update it now
    if(UPDATE == 0) 
    { // Txf("find2");
      SD_Write( LastFATSectorRead , FAT.Buffer);
      UPDATE = 1;
    } 
    // read next sector of fat
    LastFreeClusSector++; 
    // if we search the fat table until end from begining point at some where not FATAdd guess 
    //that may have some free cluster at begining of fat table so search at FATAdd.
    if( (SEARCH_COMPLETE == 0) && ( LastFreeClusSector == (FATAdd + BPB_FATSz16) ))
    { SEARCH_COMPLETE = 1;
      LastFreeClusSector = FATAdd;
    }
    LastFATSectorRead = LastFreeClusSector; 
    SD_Read( LastFATSectorRead , FAT.Buffer);
  }
  return(0); //free cluster not find
} 

//=========================================================================== 
// Hladanie subora
// Ak set = 1 tak sa filesize upravi aj vo FAT
// ak set = 0 nastavi sa len hodnota FileSize & FirstCluster. 
// Navratova hodnota 255 znamena, ze sa subor nenasiel, inac je to index najdeneho subora
//===========================================================================
unsigned char SearchFile(unsigned char *Name,unsigned int * FirstCluster,unsigned int * FileSize,unsigned char set)
{ unsigned char i, j , n ; 
  
  n = BPB_BytsPerSec / 32 ; 

  for( i = 0 ; i < RootDirSectors ; i++)
  { //for1
    SD_Read( RootDirAdd + i , Buffer1.data);
    // search 16 entry in this sector of root directory
    for( j = 0 ; j < n ; j++ )
    { //for2 
      if( Buffer1.DirEntry[ j ].DIR_Name[0] == 0x00 ) 
       return(255); // there is no more file

      if( ((Buffer1.DirEntry[ j ].DIR_Attr & 0x1f) != ATTR_LONG_NAME) && (Buffer1.DirEntry[ j ].DIR_Name[0] != 0xe5) )
      { //if1 
        if( (Buffer1.DirEntry[ j ].DIR_Attr & ( ATTR_DIRECTORY | ATTR_VOLUME_ID )) == 0x00)
        { //if2 
          //================This is a file ====================//
          Buffer1.DirEntry[ j ].DIR_Attr = '\0';
          if( !strncmp( (unsigned char *) & Buffer1.DirEntry[ j ] , Name, 11) )
          { if(set == 1)
            { Buffer1.DirEntry[ j ].DIR_FileSize = *FileSize; 
              if( *FirstCluster == 0 ) 
              { Buffer1.DirEntry[ j ].DIR_FstClusLO = 0; 
                Buffer1.DirEntry[ j ].DIR_Name[0] = 0xe5;
              }

              SD_Write(RootDirAdd + i , Buffer1.data); 
              return(j); //found the file
            } 
            else
            { *FileSize = Buffer1.DirEntry[ j ].DIR_FileSize;
              *FirstCluster = Buffer1.DirEntry[ j ].DIR_FstClusLO; 
              return(j);//found the file
            }
          }
          //============================================// 
        } //if2
      } //if1
    } //for2
  } //for1 
  return(255); // the file not found
} 

//===========================================================================
// Funkcie pre pracu so subormi
//===========================================================================
void CleanFile(unsigned char *Name)
{ unsigned int tmp , prev_clus, next_clus = 0; 

  if(FILE_NOT_FOUND == SearchFile(Name,&prev_clus,&tmp,0))
   return; // subor sa nenasiel

  tmp = 0; 
  SearchFile(Name,&tmp,&tmp,1); // clear the rootdir

  if(prev_clus == 0)
   return;  // subor je prazdny

  while(next_clus != 0xffff )
  { next_clus = GetClusEntryVal_N(prev_clus); 
    SetClusEntryVal_N(prev_clus ,0x0000);
    prev_clus = next_clus;
  } 
  UpdateFATBuffer(); 
} 

//===========================================================================
// Nastavenie casu
// ak je k dispozicii zdroj realneho casu, je mozne to prepojit
void Set_FAT_time(void)
{ short  hour, min, sec;

  hour = 0;
  min = 0;
  sec = 0;

  file_time = 0;

//  rtc_get_time(0, &hour, &min, &sec, &hsec); 
  file_time = (sec >> 1) & 0x001f; 
  file_time |= ((short)min << 5) & 0x07e0;
  file_time |= ((short)hour << 11); 
}

// Nastavenie datumu
// ak je k dispozicii zdroj realneho casu, je mozne to prepojit
void Set_FAT_Date(void)
{ short day, month, year;

  day = 1;
  month = 1;
  year = 1;
  
  file_date = 0;
//  rtc_get_date(0,&day,&month,&year); 
  file_date = day & 0x001f;
  file_date |= ((short)month << 5);
  year += 27;
  file_date |= ((short)year << 9);
}

//===========================================================================
// Vytvorenie subora
unsigned int CreateFile(unsigned char *Name)
{ unsigned char k,i, j , n ; 
  unsigned int t;
  unsigned int free_clus;

  if(!(free_clus = FindFreeCluster()))
   return(0); // find find any free cluster

  if(SearchFile(Name,&t,&t,0) != FILE_NOT_FOUND)
   return(0); //the same file name is exist

  n = BPB_BytsPerSec / 32 ; 
  // Search entire sector in root directory
  for( i = 0 ; i < RootDirSectors ; i++)
  { //for1
    SD_Read( RootDirAdd + i , Buffer1.data);
    // search 16 entry in this sector of root directory
    for( j = 0 ; j < n ; j++ )
    { //for2 
      if( (Buffer1.DirEntry[ j ].DIR_Name[0] == 0x00) || (Buffer1.DirEntry[ j ].DIR_Name[0] == 0xe5) ) 
      { for(k = 0 ; k < 11 ; k++ ) 
        { Buffer1.DirEntry[ j ].DIR_Name[k] = *Name++;
        } 
        Set_FAT_time();
        Set_FAT_Date();
        Buffer1.DirEntry[ j ].DIR_Attr = 0x20; //file attributes RSHA, Longname, Drive Label, Directory
        Buffer1.DirEntry[ j ].DIR_NTRes = 0x0; //set to zero
        Buffer1.DirEntry[ j ].DIR_CrtTimeTenth = 0x41; //creation time part in milliseconds
        Buffer1.DirEntry[ j ].DIR_CrtTime = file_time; //creation time
        Buffer1.DirEntry[ j ].DIR_CrtDate = file_date; //creation date
        Buffer1.DirEntry[ j ].DIR_LastAccDate = file_date; //last access date
        Buffer1.DirEntry[ j ].DIR_FstClusHI = 0x0; //first cluster high word 
        Buffer1.DirEntry[ j ].DIR_WrtTime = file_time; //last write time
        Buffer1.DirEntry[ j ].DIR_WrtDate = file_date; //last write date
        Buffer1.DirEntry[ j ].DIR_FstClusLO = free_clus; //first cluster low word 
        Buffer1.DirEntry[ j ].DIR_FileSize = 0x00000000; 

        SD_Write(RootDirAdd + i , Buffer1.data); 
        SetClusEntryVal_N(free_clus , 0xffff );
        return(free_clus); //find free space
      }
    } //for2
  } //for1 
  return(0); // no free cluster
} 

//=========================================================================== 
// Zapis do subora
// 1: Pouzit file_new() s nazvom subora
// 2: Opakovane pouzit file_write() kym sa nezapisu vsetky udaje
// Kazdy raz sa zapisuje 512 bajtov, len pri poslednom zapise to moze byt menej
// 3: Po zapisanie vsetkych udajov pouzit file_close() s nazvom subora ako bol pouzity pri file_new()
//=========================================================================== 
unsigned int file_size_write;
unsigned int cluster_write;
unsigned int file_step_write; 
unsigned int temp; 

//===========================================================================
// Novy subor
unsigned char file_new(unsigned char *Name)
{ unsigned char meno[11];

  // uprava mena subora na kompatibilny s FAT16
  uprav_meno(Name, meno);
  Name = meno;

  // if the file not exist so create it
  if(FILE_NOT_FOUND == SearchFile(Name,&cluster_write,&file_size_write,0))
  { if(!(cluster_write = CreateFile(Name))) 
    { file_step_write = 4; // don't do any thing
      return(1); // we didn't find any free cluster 
    } 
    // beacuse we create it now so the file is empty so the filesize equal to zero
    file_size_write = 0; 
  } 
  else 
  { file_step_write = 3; // don't do any thing
    return(2); 
  }
  // get ready for next level
  temp = 1; 
  SearchFile(Name,&temp,&file_size_write,1); 
  file_step_write = 0; 
  return(0); 
}

//=========================================================================== 
// Zapis do subora
unsigned char file_write(unsigned char * Buff , short lengh)
{ static unsigned char sector_offset = 0;
  static unsigned int first_sector_clus; 

  switch(file_step_write)
  { case 0: // when create new file 
            first_sector_clus = FirstSectorofCluster_N(cluster_write);
            SD_Write(first_sector_clus, Buff);
            file_size_write += lengh; 
            sector_offset = 1;
            file_step_write = 2;
            return(0);

    case 1: // after create new file
            temp = FindFreeCluster(); 
            // mmc is full
            if(temp == 0) 
              return(1);
            SetClusEntryVal_N(cluster_write,temp); 
            cluster_write = temp;
            first_sector_clus = FirstSectorofCluster_N(cluster_write); 
            SD_Write( first_sector_clus, Buff); 
            file_size_write += lengh;
            file_step_write = 2; 
            return(0);

    case 2: // until the clus full 
            SD_Write( first_sector_clus + sector_offset , Buff); 
            file_size_write += lengh;
            sector_offset++;
            if(sector_offset == BPB_SecPerClus)
            { file_step_write = 1;
              sector_offset = 1; 
            } 
            return(0);
  }
  return(0);
} 

//=========================================================================== 
// zatvorenie subora
void file_close(unsigned char *Name)
{ unsigned char meno[11];

  // uprava mena subora na kompatibilny s FAT16
  uprav_meno(Name, meno);
  Name = meno;

  temp = 1; 
  SearchFile(Name,&temp,&file_size_write,1); 
  UpdateFATBuffer();
}

//===========================================================================
// Citanie zo subora
// 1: Pouzit file_open() s nazvom subora (vrati 1 ak subor neexistuje, 2 ak je prazdny, 0 ak je OK)
// 2: Opakovane pouzit file_read() kym sa neprecita menej nez 512 bajtov - koniec subora
//===========================================================================
unsigned int file_size_read;
unsigned int cluster_read;
unsigned int file_step_read; 

//=========================================================================== 
// Otvorenie subora pre citanie
// vrati 1 ak subor neexistuje, 2 ak je prazdny, 0 ak je OK
unsigned char file_open(unsigned char *Name)
{ unsigned char meno[11];

  // uprava mena subora na kompatibilny s FAT16
  uprav_meno(Name, meno);
  Name = meno;

  if( FILE_NOT_FOUND == SearchFile(Name,&cluster_read,&file_size_read,0) ) 
    return(1);    // file not found
  if(cluster_read == 0) 
  { file_size_read = 0;
    return(2);    // file is empty 
  }
  file_step_read = 0;
  return(0);
}

//=========================================================================== 
// Citanie casti subora
// ak je precitana cast mensia nez 512 bajtov, znamena to koniec subora
short file_read(unsigned char * Buff)
{ static unsigned char sector_offset = 0;
  static unsigned int first_sector_clus; 

  if(file_size_read == 0)
    return(0);//total file is read or file empty
  switch(file_step_read)
  { case 0: // when open new file 
            first_sector_clus = FirstSectorofCluster_N(cluster_read);
            SD_Read(first_sector_clus, Buff);
            sector_offset = 1;
            file_step_read = 2; 
            if(file_size_read >= 512)
            { file_size_read -= 512;
              return(512);
            } 
            else
            { temp = file_size_read;
              file_size_read = 0;
              return(temp);
            }

    case 1: // after reading first clus file 
            cluster_read = GetClusEntryVal_N(cluster_read);
            first_sector_clus = FirstSectorofCluster_N(cluster_read);
            SD_Read(first_sector_clus, Buff); 
            file_step_read = 2; 
            if(file_size_read >= 512)
            { file_size_read -= 512;
              return(512);
            } 
            else
            { temp = file_size_read;
              file_size_read = 0;
              return(temp);
            }

    case 2: // until the clus fuly read 
            SD_Read( first_sector_clus + sector_offset , Buff); 
            sector_offset++;
            if(sector_offset == BPB_SecPerClus)
            { file_step_read = 1;
              sector_offset = 1; 
            } 
            if(file_size_read >= 512)
            { file_size_read -= 512; 
              return(512);
            } 
            else
            { temp = file_size_read;
              file_size_read = 0;
              return(temp);
            }
  }
  return(0);
}
