//  *****************************************************************************
//                                    main.c
// 
//        Demonstracny program pre dosku KTK ARM7
//
//        Ukladanie GPS suradnic do subora na SD karte - pouziva FAT16
//
//  *****************************************************************************

// Hlavickove subory
#include "board.h"
#include "fat.h"
#include "usart.h"
#include "sd_spi.h"
#include "retazec.h"
#include "sprintf.h"

// Globalne premenne
extern unsigned char *GPS_ok;     // smernik na buffer s prijatou spravou GPGGA
extern unsigned int GPS_sprava;   // priznak nacitania spravy

unsigned char buf1[513];          // buffer pre pracu s FAT cez SPI

// Pocet krokov casovaca na 1 us
#define STEPS   ((unsigned int) ((MCK/8) + 500000) / 1000000L)

//  *****************************************************
//  Mikrosekundove oneskorenie (max. cca 10 000)
//  *****************************************************
void delay(unsigned int us)
{
  AT91C_BASE_PMC->PMC_PCER = (1 << AT91C_ID_TC0);  // povolenie hodin pre TC0
  AT91C_BASE_TC0->TC_CMR = 1;                      // volba hodin (MCK/8)
  AT91C_BASE_TC0->TC_RC = us * STEPS;              // nastavenie porovnavacieho registra C
  AT91C_BASE_TC0->TC_CCR = AT91C_TC_CLKEN |        // povolenie hodin
                           AT91C_TC_SWTRG;         // restart casovaca

  // cakanie na ukoncenie oneskorenia
  // ked je nastaveny compare C flag, je hotovo
  while ((AT91C_BASE_TC0->TC_SR & AT91C_TC_CPCS) == 0) asm ("nop");
    
  AT91C_BASE_PMC->PMC_PCDR = (1 << AT91C_ID_TC0);  // zakazanie hodin pre TC0
}

//  *******************************************************
//  Milisekundove cakanie
//  *******************************************************
void cakaj(unsigned int ms)
{ unsigned int ii;
 
  for(ii=0; ii<ms; ii++) 
    delay(1000);
}

//  *******************************************************
//  Blikanie LED - indikacia chyby 
//  blika v seriach podla danej hodnoty
//  *******************************************************
void blikaj(int pocet)
{ int i;

  // zakazanie preruseni
  AT91C_BASE_AIC->AIC_IDCR = 0xffff;

  // zaciatocny stav
  AT91C_BASE_PIOA->PIO_CODR = D1;	      // vypnutie LED

  // blikanie    
  while(1)
  { for(i=0;i<pocet;i++)
    { AT91C_BASE_PIOA->PIO_SODR = D1;	      // zapnutie LED
      cakaj(100);
      AT91C_BASE_PIOA->PIO_CODR = D1;	      // vypnutie LED
      cakaj(100);
    }
    // medzera medzi seriami
    cakaj(400);
  }
}

//  *******************************************************
//                Hlavny program
//  *******************************************************
int main (void) 
{ int ii, pocet;
  unsigned char ret[100];  // pomocny retazec
  float dlzka, sirka;

  //******************* Uvodna inicializacia *******************

  // Povolenie Resetu - minimalna dlzka 960 us
  AT91C_BASE_RSTC->RSTC_RMR = AT91C_RSTC_URSTEN | (0x4<<8) | (unsigned int)(0xA5<<24);

  //******************* Inicializacia portu A ******************
  // Povolenie hodin pre PIOA
  AT91C_BASE_PMC->PMC_PCER = (unsigned int) 1 << AT91C_ID_PIOA; 
  // konfiguracia jednotlivych pinov
  // vystupne piny
  AT91C_BASE_PIOA->PIO_OER = D1;
  // povolenie vsetkych nakonfigurovanych pinov
  AT91C_BASE_PIOA->PIO_PER = D1;
  // vsetko je vypnute
  AT91C_BASE_PIOA->PIO_CODR = D1;

  //****************** Inicializacia periferii *****************
  // Inicializacia SPI
  SPI_init();

  // inicializacia SD karty
  if ((ii = SD_SPI_init()) != 0)
    blikaj(ii);     // indikacia chyby - treba restart

  // Inicializacai FAT
  FAT_Init(buf1);

  //******************* Hlavna cast programu *******************
  // zapnutie LED - presla uvodna faza inicializacie
  AT91C_BASE_PIOA->PIO_SODR = D1;

  // cakanie na stlacenie tlacidla S2
  while(AT91C_BASE_PIOA->PIO_PDSR & S2)     // kontrola, ci je S2 v log.1
    delay(200);                             // cakanie kvoli potlaceniu zakmitov
  AT91C_BASE_PIOA->PIO_CODR = D1;
  // cakanie na pustenie S2
  while(!(AT91C_BASE_PIOA->PIO_PDSR & S2))  // kontrola, ci je S2 v log.0
    delay(200);
  AT91C_BASE_PIOA->PIO_SODR = D1;

  // vytvorenie subora
  if (file_new("gps.xml") != 0)
    blikaj(3);                // chyba - malo miesta, alebo subor uz existuje

  // zapisanie hlavicky subora
  strcpy(buf1,"<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\" ?>\r\n");
  strcat(buf1,"<gpx xmlns=\"\" creator=\"\" version=\"1.1\">\r\n");
  strcat(buf1," <trk> <name>NAZOV</name>\r\n  <trkseg>\r\n");
  // pocet znakov v pomocnom buffri
  pocet = strlen(buf1);

  // inicializacia USART 
  USART0_init();

  // snimanie az do stlacenia tlacidla
  while(AT91C_BASE_PIOA->PIO_PDSR & S2)
  { // cakanie na spravu a jej zapis
    if (GPS_sprava == 1)
    { GPS_sprava = 0;
      // extrakcia udajov zo spravy
      // desatinna cast sirky
      sirka = 10*(GPS_ok[20]-'0')+(GPS_ok[21]-'0')+0.1*(GPS_ok[23]-'0')+0.01*(GPS_ok[24]-'0')+0.001*(GPS_ok[25]-'0')+0.0001*(GPS_ok[26]-'0');
      sirka = sirka/60.0;
      // cela cast sirky
      sirka += 10*(GPS_ok[18]-'0')+(GPS_ok[19]-'0');
      // desatinna cast dlzky
      dlzka=10*(GPS_ok[33]-'0')+(GPS_ok[34]-'0')+0.1*(GPS_ok[36]-'0')+0.01*(GPS_ok[37]-'0')+0.001*(GPS_ok[38]-'0')+0.0001*(GPS_ok[39]-'0');
      dlzka = dlzka/60.0;
      // cela cast dlzky
      dlzka += 100*(GPS_ok[30]-'0')+10*(GPS_ok[31]-'0')+(GPS_ok[32]-'0');
      rsprintf(ret,"   <trkpt lat=\"%f\" lon=\"%f\">  </trkpt>\r\n",sirka,dlzka);
      // zapis udajov do buffra 
      // zistenie volneho miesta v buffri
      if (pocet + strlen(ret) < 512)
      { strcat(buf1,ret);
        pocet += strlen(ret);
      }
      else
      { // doplnenie buffra na 512 bajtov
        strncat(buf1,ret,512-pocet);
        file_write(buf1, 512);                // zapis sektoru na kartu 
        // skopirovanie nezapisanych udajov
        strcpy(buf1,ret+512-pocet);
        pocet = strlen(buf1);
      }
    }
  }
  // zakazanie preruseni od USART
  AT91C_BASE_US0->US_IDR = 0xffff;

  // koniec hlavicky subora
  strcpy(ret,"  </trkseg>\r\n </trk>\r\n</gpx>\r\n");
  // kontrola volneho miesta
  if (strlen(buf1) < 512 - strlen(ret))
  { strcat(buf1,ret);
    file_write(buf1, strlen(buf1));
  }
  else
  { // doplnenie znakov do poctu 512
    strncat(buf1,ret,512-pocet);
    file_write(buf1, 512);
    // zapis zvysku znakov
    strcpy(buf1,ret+512-pocet);
    file_write(buf1, strlen(buf1));
  }

  // zatvorenie suboru
  file_close("gps.xml");

  // vypnutie LED - koniec merania
  AT91C_BASE_PIOA->PIO_CODR = D1;
  
  // nekonecna slucka
  while(1);
}
