/*
*
*   Podprogramy pre pracu s SD/MMC kartou cez SPI
*
*/

// Hlavickove subory
#include "board.h"
#include "sd_spi.h"

//  *******************************************************
//  Konfiguracia SPI
//  *******************************************************
void SPI_init(void)
{
  // Konfiguracia PIO radica do rezimu periferie SPI
  // PIO ma zakazany pristup na dane piny
  AT91C_BASE_PIOA->PIO_PDR = MISO | MOSI | SPCK | NPCS0;
  // povolenie periferie A
  AT91C_BASE_PIOA->PIO_ASR = MISO | MOSI | SPCK | NPCS0;

  // povolenie hodin
  AT91C_BASE_PMC->PMC_PCER = 1 << AT91C_ID_SPI;
  // reset a povolenie SPI
  AT91C_BASE_SPI->SPI_CR = AT91C_SPI_SPIEN | AT91C_SPI_SWRST;
  // konfiguracia SPI
  AT91C_BASE_SPI->SPI_MR  = AT91C_SPI_MSTR |      // MASTER mode
                            AT91C_SPI_PS_FIXED |  // pouziva sa len jedno Slave zariadenie
                            14 << 16 |            // aktivny je Chip Select 0
                            AT91C_SPI_MODFDIS;    // zakazanie detekcie Mode Fault
  // parametre prenosu
  AT91C_BASE_SPI->SPI_CSR[0] = AT91C_SPI_NCPHA |  // SPI mod 0
                               AT91C_SPI_CSAAT |  // Chip Select ostane aktivny aj po prenose 
                               AT91C_SPI_BITS_8 | // 8-bitove slova
                               SPI_SCBR;          // plna rychlost prenosu - cca 24 MHz
  // povolenie SPI
  AT91C_BASE_SPI->SPI_CR = AT91C_SPI_SPIEN;
}

//  *******************************************************
//  Vyslanie znaku na SPI a precitanie prijateho
//  *******************************************************
char SPI_RW(char znak)
{
  // cakanie na uvolnenie registra pre zapis dat
  while(!(AT91C_BASE_SPI->SPI_SR & AT91C_SPI_TDRE));
  
  // aktivacia pozadovaneho kanala
//  AT91C_BASE_SPI->SPI_MR  = AT91C_SPI_MSTR | AT91C_SPI_PS_FIXED | AT91C_SPI_MODFDIS | 0x0e << 16;
//  AT91C_BASE_SPI->SPI_CSR[0] = SPI_SCBR | AT91C_SPI_NCPHA | AT91C_SPI_CSAAT | AT91C_SPI_BITS_8;

  // vyslanie znaku
  AT91C_BASE_SPI->SPI_TDR = znak;

  // cakanie na prijem znaku (pocas vysielania znaku sa vzdy aj cita)
  while(!(AT91C_BASE_SPI->SPI_SR & AT91C_SPI_RDRF));
  // citanie znaku
  return (AT91C_BASE_SPI->SPI_RDR & 0xFF);
}

//  *******************************************************
//  Vyslanie prikazu do SD karty
//  prikaz ma vzdy 6 bajtov
//  cakanie na odpoved trva 0 az 8 bajtov
//  adresa - 32 bitova adresa bloku
//  *******************************************************
char SD_Command(char prvy, unsigned int adresa, char posledny )
{ int ii;
  char ch;

  // prazdny znak
  SPI_RW(0xff);
  // vyslanie prikazu
  SPI_RW(prvy);
  SPI_RW(adresa >> 24);             // adresa od najvyssieho bajtu
  SPI_RW((adresa >> 16) & 0xff);
  SPI_RW((adresa >> 8) & 0xff);
  SPI_RW(adresa & 0xff);
  SPI_RW(posledny);
  // caka sa na odpoved (maximalne 8 prazdnych znakov)
  for(ii=0;ii<1000;ii++)
  { ch = SPI_RW(0xff);
    if ((ch & 0x80) == 0)
      return ch;        // odpoved
  }
  // ak do 9 znakov neprisla odpoved, nastala chyba
  return 0xFF;	        // chyba
}

//  *******************************************************
//  Konfiguracia SD karty
//  Navratova hodnota: 0 - vsetko v poriadku, inac kod chyby
//  *******************************************************
int SD_SPI_init(void)
{ char ch;
  int ii, jj;

  // viacnasobna inicializacia - niekedy to neide na prvy raz
  for(jj=0;jj<5;jj++)
  { // prikaz CMD 0 - RESET do SPI modu (SanDisk MMC)
    if ((ch = SD_Command(0x40,0,0x95)) == 1)
      break;           // prikaz uspesny

    // kratke cakanie - cca 0,1 s
    for(ii=0; ii<MCK/90; ii++);
  }
  if (jj >= 5)
    return 1;        // inicializacia zlyhala

  // vysielanie CMD1 a kontrola Statusu, kym sa neukonci inicializacia karty
  ii = 0;
  do { ch = SD_Command(0x41,0,0xff);
       ii++;
       if (ii > 100000)
         return 2;    // timeout
     }
  while(ch != 0);
     
  return 0;           // vsetko v poriadku
}

//  *******************************************************
//  Nacitani bloku z SD karty (512 bajtov)
//  adresa - cislo bloku
//  Navratova hodnota: 0 - vsetko v poriadku, inac kod chyby
//  *******************************************************
int SD_Read(unsigned int adresa, unsigned char *buf)
{ int ii;
  char ch;

  // cakanie na eventualne ukoncenie predch.prikazu
  // ak karta vysiela 0x00, ma este pracu
  ii=0;
  do { ch = SPI_RW(0xff);
       ii++;
       if (ii >= 1000)
         return 2;        // timeout
     }
  while(ch != 0xff); 

  // prepocitanie cisla bloku na adresu
  adresa = adresa * 512;

  // prikaz CMD 17 - citanie bloku
  if ((ch = SD_Command(0x51,adresa, 0xff)) != 0)
    return 1;            // chyba

  // cakanie na zaciatok prenosu
  // blok udajov zacina znakom 254 (0xfe)
  do { ch = SPI_RW(0xff);
     }
  while(ch != 254);   
  
  // nacitanie dat
  for(ii=0;ii<512;ii++)
    buf[ii] = SPI_RW(0xff);

  // prenos CRC - nepouziva sa v SPI mode
  SPI_RW(0xff);
  SPI_RW(0xff);
  // prazdne znaky na zaver komunikacie
  SPI_RW(0xff);
  SPI_RW(0xff);

  return 0;         // vsetko OK
}

//  *******************************************************
//  Zapis bloku do SD karty (512 bajtov)
//  adresa - cislo bloku
//  Navratova hodnota: 0 - vsetko v poriadku, inac kod chyby
//  *******************************************************
int SD_Write(unsigned int adresa, unsigned char *buf)
{ int ii;
  char ch;

  // prepocitanie cisla bloku na adresu
  adresa = adresa << 9;

  // cakanie na eventualne ukoncenie predch.prikazu
  // ak karta vysiela 0x00, ma este pracu
  ii=0;
  do { ch = SPI_RW(0xff);
       ii++;
       if (ii >= 10000)
         return 2;        // timeout
     }
  while(ch != 0xff); 

  // prikaz CMD 24 - zapis bloku
  if (SD_Command(0x58,adresa, 0xff) != 0)
    return 1;       // chyba

  // najprv musi ist aspon jeden prazdny znak (takze pre istotu 2)
  SPI_RW(0xff);     // nic
  SPI_RW(0xff);     // nic

  // blok udajov zacina znakom 254 (0xfe)
  SPI_RW(0xfe);

  // vyslanie dat (512 B)
  for(ii=0;ii<512;ii++)
    SPI_RW(buf[ii]);

  // nakoniec dva prazdne bajty namiesto CRC
  SPI_RW(0xff);
  SPI_RW(0xff);

  // cakanie na odpoved (maximalne 8 prazdnych znakov)
  // ak boli udaje akceptovane, odpoved je 5
  for(ii=0;ii<9;ii++)
  { ch = SPI_RW(0xff);
    if ((ch & 0x1f) == 5)
      break;
  }
  if (ii >= 9)
    return 2;       // chyba zapisu
  
  // cakanie na ukoncenie zapisu
  // bez tohto cakania to nefunguje spravne ked sa pouziva FAT !
  while(SPI_RW(0xFF) != 0xFF);
    
  return 0;         // vsetko OK
}
