/*
*
*   Podprogramy pre pracu s USART
*
*/

// hlavickove subory
#include "board.h"
#include "usart.h"

// premenne pre seriovy port
unsigned char GPS_buff[2][100];        // buffer pre spravy z GPS
unsigned char *GPS_akt = GPS_buff[0],  // smernik na aktualny buffer, do ktoreho sa zapisuje sprava
              *GPS_ok;                 // smernik na buffer s prijatou spravou GPGGA
unsigned int GPS_start = 0;   // priznak startu GPS spravy
unsigned int GPS_pozicia = 0; // pozicia v buffri
unsigned int GPS_sprava = 0;  // priznak nacitania spravy

//*----------------------------------------------------------------------------
//* Prepinanie LED
//*----------------------------------------------------------------------------
void Prepni_LED(void)
{
  if (AT91C_BASE_PIOA->PIO_PDSR & D1)
    AT91C_BASE_PIOA->PIO_CODR = D1;	      // vypnutie LED
   else
    AT91C_BASE_PIOA->PIO_SODR = D1;	      // zapnutie LED
}

//*----------------------------------------------------------------------------
//* Obsluha prerusenia USART - od GPS
//*----------------------------------------------------------------------------
void USART0_prerusenie(void)
{ unsigned int znak;

  // nacitanie znaku
  znak = AT91C_BASE_US0->US_RHR; 

  // kontrola zaciatku spravy (znak $)
  if ((GPS_start == 0) && (znak == '$'))
  { GPS_start = 1;
    GPS_pozicia = 0;
  }  
    
  if (GPS_start == 1)
  { if (GPS_pozicia < 99)
    { GPS_akt[GPS_pozicia] = (char)znak;
      GPS_pozicia++;
    }
    else
    { // nieco sa pokaslalo, ideme odznova
      GPS_pozicia = 0;
      GPS_start = 0;
    }
    // kontrola konca spravy
    if (znak == 0x0A)
    { // filtracia sprav - akceptujeme len GPGGA
      if (GPS_akt[4] == 'G' && GPS_akt[5] == 'A')
      { // sprava s realnymi udajmi ma minimalne 70 znakov
        if (GPS_pozicia > 70)
        { // ukoncovaci znak retazca
          GPS_akt[GPS_pozicia] = 0;
          // prepnutie buffrov a nastavenie priznaku
          if (GPS_akt == GPS_buff[0])
          { GPS_akt = GPS_buff[1];
            GPS_ok = GPS_buff[0];
          }
          else
          { GPS_akt = GPS_buff[0];
            GPS_ok = GPS_buff[1];
          }
          GPS_sprava = 1;

          // signalizacia funkcie GPS
          Prepni_LED();
        }
      }  
      GPS_pozicia = 0;
      GPS_start = 0;
    }  
  }
  // vymazanie stavoveho bitu
  AT91C_BASE_US0->US_CR = AT91C_US_RSTSTA;
}

//*----------------------------------------------------------------------------
//* Inicializacia USART0
//*----------------------------------------------------------------------------
void USART0_init( void )
{ unsigned int pom;

  // Konfiguracia PIO radica do rezimu periferie USART
  // PIO ma zakazany pristup na dane piny
  AT91C_BASE_PIOA->PIO_PDR = AT91C_PA5_RXD0 | AT91C_PA6_TXD0;
  // povolenie periferie A
  AT91C_BASE_PIOA->PIO_ASR = AT91C_PA5_RXD0 | AT91C_PA6_TXD0;

  // povolenie hodin pre USART
  AT91C_BASE_PMC->PMC_PCER = 1 << AT91C_ID_US0;

  // konfiguracia USART
  AT91C_BASE_US0->US_IDR = 0xffff;        // zakazanie preruseni
  AT91C_BASE_US0->US_CR = AT91C_US_RSTRX | AT91C_US_RSTTX | AT91C_US_RXDIS | AT91C_US_TXDIS;

  // vypocet delitela pre danu rychlost 
  pom = ((MCK*10)/(USART0_BAUD_RATE * 16));
  if ((pom % 10) >= 5)
    pom = (pom / 10) + 1;
   else
    pom = pom / 10;

  AT91C_BASE_US0->US_BRGR = pom;        // nastavenie rychlosti
  AT91C_BASE_US0->US_TTGR = 0;          // Time Guard
  AT91C_BASE_US0->US_MR = AT91C_US_CHRL_8_BITS | 
                          AT91C_US_NBSTOP_1_BIT |
                          AT91C_US_PAR_NONE;        // udaje 8bit, 1 stop bit, ziadna parita

  // povolenie prijmu znakov
  AT91C_BASE_US0->US_CR = AT91C_US_RXEN;

  // nastavenie obsluzneho programu
  AT91C_BASE_AIC->AIC_SVR[AT91C_ID_US0] = (unsigned int) USART0_prerusenie;
  // zdroj prerusenia a priorita
  AT91C_BASE_AIC->AIC_SMR[AT91C_ID_US0] = USART0_INTERRUPT_LEVEL | AT91C_AIC_SRCTYPE_INT_HIGH_LEVEL;
  // zrusenie prerusenia (ak nejake nahodou bolo)
  AT91C_BASE_AIC->AIC_ICCR = 0x1 << AT91C_ID_US0;
  // povolenie prerusenia od USART0
  AT91C_BASE_AIC->AIC_IECR = 0x1 << AT91C_ID_US0;

  // prerusenie pri prijati znaku
  AT91C_BASE_US0->US_IER = AT91C_US_RXRDY;
}
